/**
 Demo code for "mini-threads" library in threads.s.
 Creates a few threads, and swaps back and forth between them.
 
 Orion Sky Lawlor, olawlor@acm.org, 2006/01/29 (Public Domain)
*/
#include <stdio.h>
#include <stdlib.h>

/************* mini-threads library ************/
/* A "thread": just enough state to remember what we're doing. */
typedef struct mth_thread {
   void *stackPtr; /* ESP */
   void *stackAlloc;
   int stackLen;
} mth_thread;

/* routines defined in threads.s */
void mth_swap(mth_thread *cur_th,mth_thread *new_th);
typedef void (*mth_fn)(void *fn_arg);
void mth_create(mth_thread *n,mth_fn f,void *fn_arg);

void mth_new(mth_thread *n,mth_fn f,void *fn_arg)
{
	n->stackLen=8192;
	n->stackAlloc=malloc(n->stackLen);
	/* stack pointer points to *end* of allocated region */
	n->stackPtr=n->stackLen+(char *)n->stackAlloc;
	mth_create(n,f,fn_arg);
}

/********* Code calling threads library **********/
mth_thread m,a,b;

void fn_a(void *arg) {
	int i;
	printf("Entering routine A: i==%p, arg==%p\n",&i,arg);
	for (i=0;i<10;i++) {
		printf("   A(%d)\n",i);
		mth_swap(&a,&b);
	}
	printf("Done with routine A\n");
	mth_swap(&a,&m);
}

void fn_b(void *arg) {
	int i;
	printf("Entering routine B: i==%p, arg==%p\n",&i,arg);
	for (i=0;i<10;i++) {
		printf("            B(%d)\n",i);
		mth_swap(&b,&a);
	}
	printf("Done with routine B\n");
	mth_swap(&b,&m);
}

int main() {
	printf("Started main routine. Creating threads:\n");
	mth_new(&a,fn_a,(void *)0xaaaa);
	mth_new(&b,fn_b,(void *)0xbbbb);
	printf("Swapping to threads\n");
	mth_swap(&m,&a);
	printf("Back from swap\n");
	return 0;
}
